# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from gettext import gettext as _

from django.db import transaction
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import detail_route
from rest_framework import serializers, status
from rest_framework.response import Response

from pulpcore.plugin.models import Artifact
from pulpcore.plugin.serializers import (
    AsyncOperationResponseSerializer,
    RepositoryPublishURLSerializer,
    RepositorySyncURLSerializer
)

from pulpcore.plugin.tasking import enqueue_with_reservation

from pulpcore.plugin.viewsets import (
    ContentFilter,
    ContentViewSet,
    RemoteViewSet,
    OperationPostponedResponse,
    PublisherViewSet,
)

from . import tasks
from .models import CookbookPackageContent, CookbookRemote, CookbookPublisher
from .serializers import (
    CookbookPackageContentSerializer,
    CookbookRemoteSerializer,
    CookbookPublisherSerializer)

from pulp_cookbook.metadata import CookbookMetadata


class CookbookPackageContentFilter(ContentFilter):
    """Filters for the content endpoint."""

    class Meta:
        model = CookbookPackageContent
        fields = [
            'name',
            'version'
        ]


class CookbookPackageContentViewSet(ContentViewSet):
    """The ViewSet for the content endpoint."""

    endpoint_name = 'cookbook/cookbooks'
    queryset = CookbookPackageContent.objects.all()
    serializer_class = CookbookPackageContentSerializer
    filterset_class = CookbookPackageContentFilter

    @transaction.atomic
    def create(self, request):
        data = request.data
        try:
            artifact = self.get_resource(data['artifact'], Artifact)
        except KeyError:
            raise serializers.ValidationError(detail={'artifact': _('This field is required')})

        try:
            metadata = CookbookMetadata.from_cookbook_file(artifact.file.name, data['name'])
        except KeyError:
            raise serializers.ValidationError(detail={'name': _('This field is required')})
        except FileNotFoundError:
            raise serializers.ValidationError(
                detail={'artifact': _('No metadata.json found in cookbook tar')})

        try:
            if data['version'] != metadata.version:
                raise serializers.ValidationError(
                    detail={'version': _('version does not correspond to version in cookbook tar')})
        except KeyError:
            pass
        data['version'] = metadata.version
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        content = serializer.save(dependencies=metadata.dependencies)
        content.artifact = artifact

        headers = self.get_success_headers(request.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class CookbookRemoteViewSet(RemoteViewSet):
    """The ViewSet for the remote endpoint."""

    endpoint_name = 'cookbook'
    queryset = CookbookRemote.objects.all()
    serializer_class = CookbookRemoteSerializer

    @swagger_auto_schema(operation_description="Trigger an asynchronous task to sync cookbook "
                                               "content.",
                         responses={202: AsyncOperationResponseSerializer})
    @detail_route(methods=('post',), serializer_class=RepositorySyncURLSerializer)
    def sync(self, request, pk):
        """
        Synchronizes a repository. The ``repository`` field has to be provided.
        """
        remote = self.get_object()
        serializer = RepositorySyncURLSerializer(data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)
        repository = serializer.validated_data.get('repository')
        mirror = serializer.validated_data.get('mirror', True)
        result = enqueue_with_reservation(
            tasks.synchronize,
            [repository, remote],
            kwargs={
                'remote_pk': remote.pk,
                'repository_pk': repository.pk,
                'mirror': mirror
            }
        )
        return OperationPostponedResponse(result, request)


class CookbookPublisherViewSet(PublisherViewSet):
    """The ViewSet for the publish endpoint."""

    endpoint_name = 'cookbook'
    queryset = CookbookPublisher.objects.all()
    serializer_class = CookbookPublisherSerializer

    @swagger_auto_schema(operation_description="Trigger an asynchronous task to publish "
                                               "cookbook content.",
                         responses={202: AsyncOperationResponseSerializer})
    @detail_route(methods=('post',), serializer_class=RepositoryPublishURLSerializer)
    def publish(self, request, pk):
        """
        Publishes a repository.

        Either the ``repository`` or the ``repository_version`` fields can
        be provided but not both at the same time.

        """
        publisher = self.get_object()
        serializer = RepositoryPublishURLSerializer(data=request.data,
                                                    context={'request': request})
        serializer.is_valid(raise_exception=True)
        repository_version = serializer.validated_data.get('repository_version')

        result = enqueue_with_reservation(
            tasks.publish,
            [repository_version.repository, publisher],
            kwargs={
                'publisher_pk': str(publisher.pk),
                'repository_version_pk': str(repository_version.pk)
            }
        )
        return OperationPostponedResponse(result, request)



# coding=utf-8
"""Utilities for tests for the cookbook plugin."""
from functools import partial
from unittest import SkipTest

from pulp_smash import api, selectors
from pulp_smash.pulp3.constants import REPO_PATH
from pulp_smash.pulp3.utils import (
    gen_remote,
    gen_repo,
    require_pulp_3,
    require_pulp_plugins,
    sync,
)

from pulp_cookbook.tests.functional.constants import (
    COOKBOOK_CONTENT_PATH,
    fixture_u1,
    COOKBOOK_REMOTE_PATH
)


def populate_pulp(cfg, url=None):
    """Add cookbook contents to Pulp.

    :param pulp_smash.config.PulpSmashConfig: Information about a Pulp
        application.
    :param url: The URL to a cookbook repository. Defaults
        to :data:`pulp_smash.constants.fixture_u1.url`.
    :returns: A list of dicts, where each dict describes one cookbook content in
        Pulp.
    """
    if url is None:
        url = fixture_u1.url
    client = api.Client(cfg, api.json_handler)
    remote = {}
    repo = {}
    try:
        remote.update(client.post(COOKBOOK_REMOTE_PATH, gen_remote(url)))
        repo.update(client.post(REPO_PATH, gen_repo()))
        sync(cfg, remote, repo)
    finally:
        if remote:
            client.delete(remote['_href'])
        if repo:
            client.delete(repo['_href'])
    return client.get(COOKBOOK_CONTENT_PATH)['results']


def set_up_module():
    """Skip tests Pulp 3 isn't under test or if pulp-cookbook isn't installed."""
    require_pulp_3(SkipTest)
    require_pulp_plugins({'pulp_cookbook'}, SkipTest)


skip_if = partial(selectors.skip_if, exc=SkipTest)  # pylint:disable=invalid-name
"""The ``@skip_if`` decorator, customized for unittest.

:func:`pulp_smash.selectors.skip_if` is test runner agnostic. This function is
identical, except that ``exc`` has been set to ``unittest.SkipTest``.
"""



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
from gettext import gettext as _

from rest_framework import serializers

from pulpcore.plugin.models import Artifact
from pulpcore.plugin.serializers import (
    ContentSerializer, RelatedField, RemoteSerializer, PublisherSerializer
)

from .models import CookbookPackageContent, CookbookRemote, CookbookPublisher


class CookbookPackageContentSerializer(ContentSerializer):
    """Serializer for the cookbook content."""

    name = serializers.CharField(
        help_text=_("name of the cookbook")
    )
    version = serializers.CharField(
        help_text=_("version of the cookbook"),
        required=False
    )
    dependencies = serializers.JSONField(
        help_text=_("dependencies of the cookbook"),
        read_only=True
    )
    artifact = RelatedField(
        view_name='artifacts-detail',
        help_text=_("tar archive containing the cookbook"),
        queryset=Artifact.objects.all()
    )
    artifacts = None

    class Meta:
        fields = tuple(field for field in ContentSerializer.Meta.fields if field != 'artifacts') \
            + ('name', 'version', 'dependencies', 'artifact')
        model = CookbookPackageContent


class CookbookRemoteSerializer(RemoteSerializer):
    """Serializer for the remote pointing to a universe repo."""

    cookbooks = serializers.JSONField(
        help_text=_('An optional JSON object in the format {"<cookbook name>":'
                    ' "<version_string>" }. Used to limit the cookbooks to synchronize'
                    ' from the remote'),
        required=False
    )

    class Meta:
        fields = RemoteSerializer.Meta.fields + ('cookbooks', )
        model = CookbookRemote

    # TODO: No support for version specifiers yet, only cookbook names
    def validate_cookbooks(self, value):
        if value == "":  # blank value
            return value
        if isinstance(value, dict):
            if all(value.keys()):
                return value
        raise serializers.ValidationError(
            _('Format must be {"<cookbook_name>" : "version_string" }')
        )


class CookbookPublisherSerializer(PublisherSerializer):
    """Serializer for the publisher."""

    class Meta:
        fields = PublisherSerializer.Meta.fields
        model = CookbookPublisher



# coding=utf-8
"""Tests that sync cookbook plugin repositories."""
import unittest

from pulp_smash import api, config
from pulp_smash.exceptions import TaskReportError
from pulp_smash.pulp3.constants import DOWNLOAD_POLICIES as ALL_DOWNLOAD_POLICIES
from pulp_smash.pulp3.constants import REPO_PATH
from pulp_smash.pulp3.utils import (
    delete_orphans,
    gen_remote,
    gen_repo,
    get_added_content,
    get_content,
    get_removed_content,
    sync,
)

from pulp_cookbook.tests.functional.constants import (
    fixture_u1,
    COOKBOOK_REMOTE_PATH,
    DOWNLOAD_POLICIES
)
from pulp_cookbook.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class SyncCookbookRepoTestCase(unittest.TestCase):
    """Sync repositories with the cookbook plugin."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()

    def verify_counts(self, repo, all_count, added_count, removed_count):
        self.assertEqual(len(get_content(repo)['cookbook']), all_count)
        self.assertEqual(len(get_added_content(repo)['cookbook']), added_count)
        self.assertEqual(len(get_removed_content(repo)['cookbook']), removed_count)

    def sync_and_inspect_task_report(self, remote, repo, download_count, mirror=None):
        """Do a sync and verify the number of downloaded artifacts.

        Returns:
            Task report structure

        """
        if mirror is None:
            sync_resp = sync(self.cfg, remote, repo)
        else:
            sync_resp = sync(self.cfg, remote, repo, mirror=mirror)
        tasks = tuple(api.poll_spawned_tasks(self.cfg, sync_resp))
        self.assertEqual(len(tasks), 1)
        for report in tasks[0]['progress_reports']:
            if report['message'] == "Downloading Artifacts":
                self.assertEqual(report['done'], download_count)
                break
        else:
            self.fail("Could not find 'Downloading Artifacts' stage in task report")
        return tasks[0]

    def test_sync(self):
        """Sync repositories with the cookbook plugin.

        In order to sync a repository a remote has to be associated within
        this repository. When a repository is created this version field is set
        as None. After a sync the repository version is updated.

        Do the following:

        1.  Delete orphan content units, create a repository, and an remote.
        2.  Assert that repository version is None
        3.  Sync the remote.
        4.  Assert that repository version is not None and the artifact/content counts
        5.  Sync the remote one more time.
        6.  Assert that repository version is different from the previous one. Assert that no
            artifact was downloaded.
        7.  Add a filter for a single cookbook name to the remote.
        8.  Sync the remote.
        9.  Assert artifact/content counts
        10. Change the filter for a single cookbook name in the remote.
        11. Sync the remote with "mirror=False".
        12. Assert artifact/content counts (content must have been added)
        """
        delete_orphans(self.cfg)

        client = api.Client(self.cfg, api.json_handler)

        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo['_href'])

        body = gen_remote(fixture_u1.url)
        remote = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote['_href'])

        # Sync the full repository.
        self.assertIsNone(repo['_latest_version_href'])

        all_cookbook_count = fixture_u1.cookbook_count()
        task = self.sync_and_inspect_task_report(remote, repo, all_cookbook_count)

        repo = client.get(repo['_href'])
        latest_version_href = repo['_latest_version_href']
        self.assertIsNotNone(latest_version_href)
        self.assertEqual(latest_version_href,
                         task['created_resources'][0])
        self.verify_counts(repo,
                           all_cookbook_count, all_cookbook_count, 0)

        # Sync the full repository again.
        self.sync_and_inspect_task_report(remote, repo, 0)
        repo = client.get(repo['_href'])
        self.assertNotEqual(latest_version_href, repo['_latest_version_href'])
        self.verify_counts(repo, all_cookbook_count, 0, 0)

        # Sync the repository with a filter (mirror mode is the default).
        client.patch(remote['_href'], {'cookbooks': {fixture_u1.example1_name: ''}})
        self.sync_and_inspect_task_report(remote, repo, 0)
        repo = client.get(repo['_href'])
        self.assertNotEqual(latest_version_href, repo['_latest_version_href'])
        example1_count = fixture_u1.cookbook_count([fixture_u1.example1_name])
        self.verify_counts(repo,
                           example1_count,
                           0,
                           all_cookbook_count - example1_count)

        # Sync the repository with another filter and add cookbooks (mirror=False).
        client.patch(remote['_href'], {'cookbooks': {fixture_u1.example2_name: ''}})
        self.sync_and_inspect_task_report(remote, repo, 0, mirror=False)
        repo = client.get(repo['_href'])
        self.assertNotEqual(latest_version_href, repo['_latest_version_href'])
        self.verify_counts(repo,
                           fixture_u1.cookbook_count([fixture_u1.example1_name,
                                                      fixture_u1.example2_name]),
                           fixture_u1.cookbook_count([fixture_u1.example2_name]),
                           0)


class SyncInvalidTestCase(unittest.TestCase):
    """Sync a repository with an invalid given url on the remote."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)

    def test_invalid_url(self):
        """Sync a repository using a remote url that does not exist.

        Test that we get a task failure. See :meth:`do_test`.
        """
        context = self.do_test('http://example.com/invalid/')
        self.assertIsNotNone(context.exception.task['error']['description'])

    def test_invalid_policies(self):
        """Sync a repository using all unsupported policies.

        Test that we get a task failure when using an unsupported policy. See
        :meth:`do_test`.
        """
        for policy in ALL_DOWNLOAD_POLICIES:
            if policy not in DOWNLOAD_POLICIES:
                context = self.do_test(fixture_u1.url, policy=policy)
                self.assertIsNotNone(context.exception.task['error']['description'])

    def do_test(self, url, **remote_kwargs):
        """Sync a repository given ``url`` on the remote."""
        repo = self.client.post(REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo['_href'])
        body = gen_remote(url=url, **remote_kwargs)
        remote = self.client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(self.client.delete, remote['_href'])
        with self.assertRaises(TaskReportError) as context:
            sync(self.cfg, remote, repo)
        return context



# coding=utf-8
"""Utilities for file plugin tests."""
from pulp_smash import utils
from pulp_smash.pulp3.utils import get_content


def gen_publisher(**kwargs):
    """Return a semi-random dict for use in creating a publisher."""
    data = {'name': utils.uuid4()}
    data.update(kwargs)
    return data


def get_content_and_unit_paths(repo):
    """Return the publication relative path of content units present in a cookbook repository.

    :param repo: A dict of information about the repository.
    :returns: A list of tuples (content unit, publication path of the unit) present in a given
              repository.
    """
    # The "relative_path" is actually a file path and name
    def rel_path(content):
        path_format = 'cookbook_files/{0}/{2}/{0}-{1}.tar.gz'
        return path_format.format(content['name'],
                                  content['version'],
                                  content['version'].replace('.', '_'))
    content = get_content(repo)
    assert list(content.keys()) == ['cookbook']
    return [(content_unit, rel_path(content_unit)) for content_unit in content['cookbook']]



# coding=utf-8

from typing import Dict, Iterable, List, Optional
from urllib.parse import urljoin

from pulp_smash.pulp3.constants import (
    BASE_PUBLISHER_PATH,
    BASE_REMOTE_PATH,
    CONTENT_PATH
)

DOWNLOAD_POLICIES = ['immediate']

FIXTURES_BASE_URL = 'https://pulpcookbook.s3.wasabisys.com/'

COOKBOOK_CONTENT_PATH = urljoin(CONTENT_PATH, 'cookbook/cookbooks/')

COOKBOOK_REMOTE_PATH = urljoin(BASE_REMOTE_PATH, 'cookbook/')

COOKBOOK_PUBLISHER_PATH = urljoin(BASE_PUBLISHER_PATH, 'cookbook/')

COOKBOOK_BASE_LIVE_API_PATH = '/pulp_cookbook/market/'


class CookbookFixture:
    """Class holding information about a universe cookbook repo fixture."""

    def __init__(self, repo_name: str, cookbooks: Dict[str, List[str]]) -> None:
        self.url = urljoin(FIXTURES_BASE_URL, repo_name + '/')
        self.cookbooks = cookbooks

    def cookbook_count(self, names: Optional[Iterable] = None) -> int:
        """Return the number of cookbooks or the number of cookbooks with a specific name."""
        if not names:
            names = self.cookbooks.keys()
        return sum(len(self.cookbooks[name]) for name in names)

    def cookbook_download_url(self, name: str, version: str):
        uversion = version.replace('.', '_')
        print("kjk", self.url)
        return urljoin(self.url, f"cookbook_files/{name}/{uversion}/{name}-{version}.tar.gz")


fixture_u1 = CookbookFixture('u1', {
    "apache2": ["5.0.1", "5.0.0", "3.0.1", "1.10.2", "1.9.6", "1.9.1"],
    "apt": ["7.0.0", "6.0.1", "5.0.1"],
    "ubuntu": ["2.0.1", "1.0.0"],
    "nginx": ["8.0.0", "2.7.4", "2.7.2", "2.7.0", "2.6.0", "2.5.0"]
})
fixture_u1.example1_name = 'ubuntu'
fixture_u1.example1_version = '2.0.1'
fixture_u1.example2_name = 'nginx'
fixture_u1.example2_version = '2.7.4'

COOKBOOK2_FIXTURE_URL = urljoin(FIXTURES_BASE_URL, 'u2/')
"""The URL to a cookbook universe repository."""

COOKBOOK2_URL = urljoin(COOKBOOK2_FIXTURE_URL, '1.iso')
"""The URL to an ISO cookbook at :data:`COOKBOOK2_FIXTURE_URL`."""



# coding=utf-8
"""Tests that publish cookbook plugin repositories."""
import unittest
from random import choice
from urllib.parse import urljoin

from requests.exceptions import HTTPError

from pulp_smash import api, config
from pulp_smash.pulp3.constants import REPO_PATH
from pulp_smash.pulp3.utils import (
    gen_remote,
    gen_repo,
    get_versions,
    publish,
    sync,
)

from pulp_cookbook.tests.functional.api.utils import gen_publisher
from pulp_cookbook.tests.functional.constants import (
    COOKBOOK_CONTENT_PATH,
    fixture_u1,
    COOKBOOK_REMOTE_PATH,
    COOKBOOK_PUBLISHER_PATH,
)
from pulp_cookbook.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class PublishAnyRepoVersionTestCase(unittest.TestCase):
    """Test whether a particular repository version can be published.

    This test targets the following issues:

    * `Pulp #3324 <https://pulp.plan.io/issues/3324>`_
    * `Pulp Smash #897 <https://github.com/PulpQE/pulp-smash/issues/897>`_
    """

    def test_all(self):
        """Test whether a particular repository version can be published.

        1. Create a repository with at least 2 repository versions.
        2. Create a publication by supplying the latest ``repository_version``.
        3. Assert that the publication ``repository_version`` attribute points
           to the latest repository version.
        4. Create a publication by supplying the non-latest
           ``repository_version``.
        5. Assert that the publication ``repository_version`` attribute points
           to the supplied repository version.
        6. Assert that an exception is raised when providing two different
           repository versions to be published at same time.
        """
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)
        body = gen_remote(fixture_u1.url)
        remote = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote['_href'])

        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo['_href'])

        sync(cfg, remote, repo)

        publisher = client.post(COOKBOOK_PUBLISHER_PATH, gen_publisher())
        self.addCleanup(client.delete, publisher['_href'])

        # Step 1
        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo['_href'])
        for cookbook_content in client.get(COOKBOOK_CONTENT_PATH)['results']:
            client.post(
                repo['_versions_href'],
                {'add_content_units': [cookbook_content['_href']]}
            )
        version_hrefs = tuple(ver['_href'] for ver in get_versions(repo))
        non_latest = choice(version_hrefs[:-1])

        # Step 2
        publication = publish(cfg, publisher, repo)

        # Step 3
        self.assertEqual(publication['repository_version'], version_hrefs[-1])

        # Step 4
        publication = publish(cfg, publisher, repo, non_latest)

        # Step 5
        self.assertEqual(publication['repository_version'], non_latest)

        # Step 6
        with self.assertRaises(HTTPError):
            body = {'repository': repo['_href'],
                    'repository_version': non_latest}
            client.post(urljoin(publisher['_href'], 'publish/'), body)



# coding=utf-8
"""Tests that CRUD publishers."""
import unittest

from requests.exceptions import HTTPError

from pulp_smash import api, config
from pulp_smash.pulp3.constants import REPO_PATH
from pulp_smash.pulp3.utils import gen_repo

from pulp_cookbook.tests.functional.api.utils import gen_publisher
from pulp_cookbook.tests.functional.constants import COOKBOOK_PUBLISHER_PATH
from pulp_cookbook.tests.functional.utils import set_up_module as setUpModule  # noqa:F401
from pulp_cookbook.tests.functional.utils import skip_if


class CRUDPublishersTestCase(unittest.TestCase):
    """CRUD publishers."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables.

        In order to create a publisher a repository has to be created first.
        """
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.publisher = {}
        cls.repo = cls.client.post(REPO_PATH, gen_repo())

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        cls.client.delete(cls.repo['_href'])

    def test_01_create_publisher(self):
        """Create a publisher."""
        body = gen_publisher()
        type(self).publisher = self.client.post(COOKBOOK_PUBLISHER_PATH, body)
        for key, val in body.items():
            with self.subTest(key=key):
                self.assertEqual(self.publisher[key], val)

    @skip_if(bool, 'publisher', False)
    def test_02_create_same_name(self):
        """Try to create a second publisher with an identical name.

        See: `Pulp Smash #1055
        <https://github.com/PulpQE/pulp-smash/issues/1055>`_.
        """
        body = gen_publisher()
        body['name'] = self.publisher['name']
        with self.assertRaises(HTTPError):
            self.client.post(COOKBOOK_PUBLISHER_PATH, body)

    @skip_if(bool, 'publisher', False)
    def test_02_read_publisher(self):
        """Read a publisher by its href."""
        publisher = self.client.get(self.publisher['_href'])
        for key, val in self.publisher.items():
            with self.subTest(key=key):
                self.assertEqual(publisher[key], val)

    @skip_if(bool, 'publisher', False)
    def test_02_read_publishers(self):
        """Read a publisher by its name."""
        page = self.client.get(COOKBOOK_PUBLISHER_PATH, params={
            'name': self.publisher['name']
        })
        self.assertEqual(len(page['results']), 1)
        for key, val in self.publisher.items():
            with self.subTest(key=key):
                self.assertEqual(page['results'][0][key], val)

    @skip_if(bool, 'publisher', False)
    def test_03_partially_update(self):
        """Update a publisher using HTTP PATCH."""
        body = gen_publisher()
        self.client.patch(self.publisher['_href'], body)
        type(self).publisher = self.client.get(self.publisher['_href'])
        for key, val in body.items():
            with self.subTest(key=key):
                self.assertEqual(self.publisher[key], val)

    @skip_if(bool, 'publisher', False)
    def test_04_fully_update(self):
        """Update a publisher using HTTP PUT."""
        body = gen_publisher()
        self.client.put(self.publisher['_href'], body)
        type(self).publisher = self.client.get(self.publisher['_href'])
        for key, val in body.items():
            with self.subTest(key=key):
                self.assertEqual(self.publisher[key], val)

    @skip_if(bool, 'publisher', False)
    def test_05_delete(self):
        """Delete a publisher."""
        self.client.delete(self.publisher['_href'])
        with self.assertRaises(HTTPError):
            self.client.get(self.publisher['_href'])



# coding=utf-8
"""Tests that perform actions over content unit."""
import unittest

from requests.exceptions import HTTPError

from pulp_smash import api, config, utils
from pulp_smash.pulp3.constants import ARTIFACTS_PATH
from pulp_smash.pulp3.utils import delete_orphans

from pulp_cookbook.tests.functional.constants import fixture_u1, COOKBOOK_CONTENT_PATH
from pulp_cookbook.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


def _gen_content_unit_attrs(artifact, name):
    """Generate a dict with content unit attributes.

    :param: artifact: A dict of info about the artifact.
    :returns: A dict for use in creating a content unit.
    """
    return {'artifact': artifact['_href'], 'name': name}


class ContentUnitTestCase(unittest.TestCase):
    """CRUD content unit.

    This test targets the following issues:

    * `Pulp #2872 <https://pulp.plan.io/issues/2872>`_
    * `Pulp #3445 <https://pulp.plan.io/issues/3445>`_
    * `Pulp Smash #870 <https://github.com/PulpQE/pulp-smash/issues/870>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variable."""
        cls.cfg = config.get_config()
        delete_orphans(cls.cfg)
        cls.content_unit = {}
        cls.client = api.Client(cls.cfg, api.json_handler)
        download_url = fixture_u1.cookbook_download_url(fixture_u1.example1_name,
                                                        fixture_u1.example1_version)
        cookbooks = {'file': utils.http_get(download_url)}
        cls.artifact = cls.client.post(ARTIFACTS_PATH, files=cookbooks)

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        delete_orphans(cls.cfg)

    def test_01_create_content_unit(self):
        """Create content unit."""
        attrs = _gen_content_unit_attrs(self.artifact, fixture_u1.example1_name)
        self.content_unit.update(self.client.post(COOKBOOK_CONTENT_PATH, attrs))
        for key, val in attrs.items():
            with self.subTest(key=key):
                self.assertEqual(self.content_unit[key], val)

    def test_02_read_content_unit(self):
        """Read a content unit by its href."""
        content_unit = self.client.get(self.content_unit['_href'])
        for key, val in self.content_unit.items():
            with self.subTest(key=key):
                self.assertEqual(content_unit[key], val)

    def test_02_read_content_units(self):
        """Read a content unit by its name and version."""
        page = self.client.get(COOKBOOK_CONTENT_PATH, params={
            'name': self.content_unit['name'],
            'version': self.content_unit['version'],
        })
        self.assertEqual(len(page['results']), 1)
        for key, val in self.content_unit.items():
            with self.subTest(key=key):
                self.assertEqual(page['results'][0][key], val)

    def test_03_partially_update(self):
        """Attempt to update a content unit using HTTP PATCH.

        This HTTP method is not supported and a HTTP exception is expected.
        """
        attrs = _gen_content_unit_attrs(self.artifact, fixture_u1.example1_name)
        with self.assertRaises(HTTPError) as exc:
            self.client.patch(self.content_unit['_href'], attrs)
        self.assertEqual(exc.exception.response.status_code, 405)

    def test_03_fully_update(self):
        """Attempt to update a content unit using HTTP PUT.

        This HTTP method is not supported and a HTTP exception is expected.
        """
        attrs = _gen_content_unit_attrs(self.artifact, fixture_u1.example1_name)
        with self.assertRaises(HTTPError) as exc:
            self.client.put(self.content_unit['_href'], attrs)
        self.assertEqual(exc.exception.response.status_code, 405)

    def test_04_delete(self):
        """Attempt to delete a content unit using HTTP DELETE.

        This HTTP method is not supported and a HTTP exception is expected.
        """
        with self.assertRaises(HTTPError) as exc:
            self.client.delete(self.content_unit['_href'])
        self.assertEqual(exc.exception.response.status_code, 405)



import logging

from gettext import gettext as _
from urllib.parse import urljoin

from pulpcore.plugin.models import Artifact, ProgressBar, Remote, Repository
from pulpcore.plugin.stages import (
    DeclarativeArtifact, DeclarativeContent, DeclarativeVersion, Stage,
    ArtifactDownloader, ArtifactSaver,
    QueryExistingContentUnits, ContentUnitSaver
)

from pulp_cookbook.app.models import CookbookPackageContent, CookbookRemote
from pulp_cookbook.metadata import Universe


log = logging.getLogger(__name__)


class ExistingContentNeedsNoArtifacts(Stage):
    """
    A Stages API stage that removes artifacts that are known (i.e. do not need to be downloaded).

    A Stages API stage that removes all
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` instances from
    :class:`~pulpcore.plugin.stages.DeclarativeContent` units if the respective
    :class:`~pulpcore.plugin.models.Content` is already existing.

    """

    async def __call__(self, in_q, out_q):
        """
        The coroutine for this stage.

        Args:
            in_q (:class:`asyncio.Queue`): The queue to receive
                :class:`~pulpcore.plugin.stages.DeclarativeContent` objects from.
            out_q (:class:`asyncio.Queue`): The queue to put
                :class:`~pulpcore.plugin.stages.DeclarativeContent` into.

        Returns:
            The coroutine for this stage.

        """
        async for batch in self.batches(in_q):
            for declarative_content in batch:
                if declarative_content.content.pk is not None:
                    declarative_content.d_artifacts = []
                await out_q.put(declarative_content)
        await out_q.put(None)


class CookbookFirstStage(Stage):
    """The first stage of the pulp_cookbook sync pipeline."""

    def __init__(self, remote):
        """
        The first stage of the pulp_cookbook sync pipeline.

        Get the `universe` of the remote repo, parse it and inject a
        DeclarativeContent instance for each cookbook found.

        Args:
            remote (CookbookRemote): The remote data to be used when syncing
        """
        self.remote = remote

    async def __call__(self, in_q, out_q):
        """
        Build and emit `DeclarativeContent` from the Manifest data.

        If a cookbook specifier is set in the remote, cookbooks are filtered
        using this specifier.

        Args: in_q (asyncio.Queue): Unused because the first stage doesn't read
            from an input queue. out_q (asyncio.Queue): The out_q to send
            `DeclarativeContent` objects to

        """
        with ProgressBar(message='Downloading Metadata', total=1) as pb:
            downloader = self.remote.get_downloader(url=urljoin(self.remote.url + '/', 'universe'))
            result = await downloader.run()
            pb.increment()

        cookbook_names = self.remote.specifier_cookbook_names()

        with ProgressBar(message='Parsing Metadata') as pb:
            universe = Universe(result.path)
            for entry in universe.read():
                if cookbook_names and entry.name not in cookbook_names:
                    continue
                cookbook = CookbookPackageContent(name=entry.name, version=entry.version,
                                                  dependencies=entry.dependencies)
                artifact = Artifact()
                da = DeclarativeArtifact(artifact, entry.download_url,
                                         cookbook.relative_path(), self.remote)
                dc = DeclarativeContent(content=cookbook, d_artifacts=[da])
                pb.increment()
                await out_q.put(dc)
        await out_q.put(None)


class CookbookDeclarativeVersion(DeclarativeVersion):
    """Implement pulp_cookbook's stape API pipeline."""

    def pipeline_stages(self, new_version):
        return [
            self.first_stage,
            QueryExistingContentUnits(), ExistingContentNeedsNoArtifacts(),
            ArtifactDownloader(), ArtifactSaver(),
            ContentUnitSaver()
        ]


def synchronize(remote_pk, repository_pk, mirror):
    """
    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.

    Raises:
        ValueError: When url is empty.

    """
    remote = CookbookRemote.objects.get(pk=remote_pk)
    repository = Repository.objects.get(pk=repository_pk)
    if not remote.url:
        raise ValueError(_('A remote must have a url specified to synchronize.'))

    if remote.policy != Remote.IMMEDIATE:
        raise ValueError(_('Sync supports "{}" policy only').format(Remote.IMMEDIATE))

    first_stage = CookbookFirstStage(remote)
    CookbookDeclarativeVersion(first_stage, repository, mirror).create()



# coding=utf-8
"""Tests that verify download of content served by Pulp."""
import hashlib
import unittest
from random import choice
from urllib.parse import urljoin

from pulp_smash import api, config, utils
from pulp_smash.pulp3.constants import DISTRIBUTION_PATH, REPO_PATH
from pulp_smash.pulp3.utils import (
    gen_distribution,
    gen_remote,
    gen_repo,
    publish,
    sync,
)

from pulp_cookbook.tests.functional.api.utils import (
    gen_publisher,
    get_content_and_unit_paths,
)
from pulp_cookbook.tests.functional.constants import (
    fixture_u1,
    COOKBOOK_REMOTE_PATH,
    COOKBOOK_PUBLISHER_PATH,
    COOKBOOK_BASE_LIVE_API_PATH,
)
from pulp_cookbook.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class DownloadContentTestCase(unittest.TestCase):
    """Verify whether content served by pulp can be downloaded."""

    def test_all(self):
        """Verify whether content served by pulp can be downloaded.

        The process of publishing content is more involved in Pulp 3 than it
        was under Pulp 2. Given a repository, the process is as follows:

        1. Create a publication from the repository. (The latest repository
           version is selected if no version is specified.) A publication is a
           repository version plus metadata.
        2. Create a distribution from the publication. The distribution defines
           at which URLs a publication is available. For the cookbook plugin
           this is below a live API at ``pulp_cookbook/market/`` e.g.
           ``http://example.com/pulp_cookbook/market/foo/`` and
           ``http://example.com/pulp_cookbook/market/bar/``.

        Do the following:

        1. Create, populate, publish, and distribute a repository.
        2. Select a random content unit in the distribution. Download that
           content unit from Pulp, and verify that the content unit has the
           same checksum when fetched directly from fixtures.

        This test targets the following issues:

        * `Pulp #2895 <https://pulp.plan.io/issues/2895>`_
        * `Pulp Smash #872 <https://github.com/PulpQE/pulp-smash/issues/872>`_
        """
        cfg = config.get_config()
        client = api.Client(cfg, api.json_handler)

        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo['_href'])

        body = gen_remote(fixture_u1.url)
        remote = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote['_href'])

        sync(cfg, remote, repo)
        repo = client.get(repo['_href'])

        # Create a publisher.
        publisher = client.post(COOKBOOK_PUBLISHER_PATH, gen_publisher())
        self.addCleanup(client.delete, publisher['_href'])

        # Create a publication.
        publication = publish(cfg, publisher, repo)
        self.addCleanup(client.delete, publication['_href'])

        # Create a distribution.
        body = gen_distribution()
        body['publication'] = publication['_href']
        distribution = client.post(DISTRIBUTION_PATH, body)
        self.addCleanup(client.delete, distribution['_href'])

        # pulp_cookbook universe live endpoint contains
        # all cookbooks
        distribution_base_url = cfg.get_hosts('api')[0].roles['api']['scheme']
        distribution_base_url += '://' + distribution['base_url'] + '/'

        api_host_cfg = cfg.get_hosts('api')[0]
        universe_url = api_host_cfg.roles['api']['scheme']
        universe_url += '://' + api_host_cfg.hostname
        universe_url += ':{}'.format(api_host_cfg.roles['api']['port'])
        universe_url += COOKBOOK_BASE_LIVE_API_PATH
        universe_url += distribution['base_path'] + '/universe'

        universe = client.get(universe_url)

        for content, publication_path in get_content_and_unit_paths(repo):
            u_url = universe[content['name']][content['version']]['download_url']
            self.assertEqual(u_url, urljoin(distribution_base_url, publication_path))

        # Pick a cookbook, and download it from both Fixtures…
        _, unit_path = choice(get_content_and_unit_paths(repo))
        fixtures_hash = hashlib.sha256(
            utils.http_get(urljoin(fixture_u1.url, unit_path))
        ).hexdigest()

        # …and Pulp content
        client.response_handler = api.safe_handler

        unit_url = urljoin(distribution_base_url, unit_path)

        pulp_hash = hashlib.sha256(client.get(unit_url).content).hexdigest()
        self.assertEqual(fixtures_hash, pulp_hash)
